"""
Functionality for server-side ibis transforms of vega charts.
"""
import json
import re
import typing
import concurrent.futures
import ibis.client
import ibis.omniscidb

import altair
import altair.vegalite.v3.display
import opentracing

from .core import apply
from .globals import _expr_map
from .tracer import tracer

__all__ = ["query_target_func"]

executor = concurrent.futures.ThreadPoolExecutor()


# We need a new connection per thread, which is too memory intensive
# Should just create one per threadpool and then reuse
# But for now we just disable

ENABLE_MULTIPROCESSING = False


def execute_new_client(expr):
    """
    Execute with new connection b/c connections are not threadsafe
    """
    (backend,) = list(ibis.client.find_backends(expr))
    assert isinstance(backend, ibis.omniscidb.OmniSciDBClient)
    with tracer.start_span("ibis:execute:new-client") as execute_span:
        new_client = ibis.omniscidb.OmniSciDBClient(
            uri=backend.uri,
            host=backend.host,
            port=backend.port,
            user=backend.user,
            protocol=backend.protocol,
            session_id=backend.session_id,
            database=backend.db_name,
            password=backend.password,
        )
    with new_client:
        return new_client.execute(expr)

    # reset db name so it reconnects


def query_target_func(comm, msg):
    """
    Target function for actually evaluating the `queryibis` transform.
    """
    # These are the paramaters passed to the vega transform
    parameters: dict = msg["content"]["data"]

    def callback(future):
        comm.send(future.result())

    if ENABLE_MULTIPROCESSING:
        executor.submit(execute_query, parameters).add_done_callback(callback)
    else:
        comm.send(execute_query(parameters))


def execute_query(parameters: dict):
    injected_span: object = parameters.pop("span")
    with tracer.start_active_span(
        "queryibis",
        references=[
            opentracing.child_of(
                (tracer.extract(opentracing.Format.TEXT_MAP, injected_span))
            )
        ],
    ) as scope:
        scope.span.log_kv(parameters)
        name: str = parameters.pop("name")
        transforms: typing.Optional[str] = parameters.pop("transform", None)

        if name not in _expr_map:
            raise ValueError(f"{name} is not an expression known to us!")
        expr = _expr_map[name]
        scope.span.log_kv({"sql:initial": expr.compile()})
        if transforms:
            # Replace all string instances of data references with value in schema
            for k, v in parameters.items():
                # All data items are added to parameters as `:<data name>`.
                # They also should  be in the `data` paramater, but you have to call
                # this with a tuple which I am not sure where to get from
                # https://github.com/vega/vega/blob/65fe7cb2485be90e16298d9dff87bf56045afb8d/packages/vega-transforms/src/Filter.js#L48
                if not k.startswith(":"):
                    continue
                k = k[1:]
                res = json.dumps(v)
                for t in transforms:
                    if t["type"] == "filter" or t["type"] == "formula":
                        t["expr"] = _patch_vegaexpr(t["expr"], k, res)
            try:
                expr = apply(expr, transforms)
            except Exception as e:
                raise ValueError(
                    f"Failed to convert {transforms} with error message message '{e}'"
                )
        with tracer.start_span("ibis:execute") as execute_span:
            execute_span.log_kv({"sql": expr.compile()})
            if ENABLE_MULTIPROCESSING:
                data = execute_new_client(expr)
            else:
                data = expr.execute()
        return altair.to_values(data)["values"]


def _patch_vegaexpr(expr: str, name: str, value: str) -> str:
    quote = "(['\"])"
    expr = re.sub(f"data\({quote}{name}{quote}\)", value, expr)
    expr = re.sub(
        f"vlSelectionTest\({quote}{name}{quote}", f"vlSelectionTest({value}", expr
    )
    return expr



import ibis


def timeunit(transform: dict, expr: ibis.Expr) -> ibis.Expr:
    """
    Apply a vega time unit transform to an ibis expression.
    https://vega.github.io/vega/docs/transforms/timeunit/

    It transforms it into the Ibis truncate expression.
    https://docs.ibis-project.org/generated/ibis.expr.api.TimestampValue.truncate.html

    Parameters
    ----------
    transform: dict
        A JSON-able dictionary representing the vega transform.
    expr: ibis.Expr
        The expression to which to apply the transform.

    Returns
    -------
    transformed_expr: the transformed expression
    """
    assert transform.pop("type") == "timeunit"
    field = expr[transform.pop("field")]
    as_start, as_end = transform.pop("as")
    units = transform.pop("units")
    if transform:
        raise NotImplementedError(
            f"timeunit transform: {list(transform)} keys are not supported"
        )
    if units == ["year"]:
        start = field.truncate("Y")
        delta = ibis.interval(years=1)
    elif units == ["year", "month"]:
        start = field.truncate("M")
        delta = ibis.interval(months=1)
    elif units == ["year", "month", "date"]:
        start = field.truncate("D")
        delta = ibis.interval(days=1)
    elif units == ["year", "month", "date", "hours"]:
        start = field.truncate("h")
        delta = ibis.interval(hours=1)
    elif units == ["year", "month", "date", "hours", "minutes"]:
        start = field.truncate("m")
        delta = ibis.interval(minutes=1)
    elif units == ["year", "month", "date", "hours", "minutes", "seconds"]:
        start = field.truncate("s")
        delta = ibis.interval(seconds=1)
    elif units == [
        "year",
        "month",
        "date",
        "hours",
        "minutes",
        "seconds",
        "milliseconds",
    ]:
        start = field.truncate("ms")
        delta = ibis.interval(milliseconds=1)
    else:
        raise NotImplementedError(
            f"timeunit transform: {units} units are not supported"
        )
    return expr.mutate([start.name(as_start), (start + delta).name(as_end)])



"""Setup script for Ibis Vega Transform."""

# Standard library imports
import ast
import os

# Third party imports
import setuptools

# Constants
HERE = os.path.abspath(os.path.dirname(__file__))


def get_version(module="ibis_vega_transform"):
    """Get version."""
    data = read(os.path.join(HERE, module, "__init__.py"))
    lines = data.split("\n")
    for line in lines:
        if line.startswith("__version__"):
            version = ast.literal_eval(line.split("=")[-1].strip())
            return version


def read(path, encoding="utf-8"):
    """
    Read file specified by path relative to file and return content.
    """
    path = os.path.join(HERE, path)

    content = ""
    if os.path.exists(path):
        with open(path, encoding=encoding) as fh:
            content = fh.read()

    return content


setuptools.setup(
    name="ibis-vega-transform",
    version=get_version(),
    url="https://github.com/Quansight/ibis-vega-transform",
    author="Ian Rose and Saul Shanabrook",
    author_email="ian.rose@quansight.com",
    license="Apache-2.0 license",
    description="Evaluate Vega transforms using Ibis expressions",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "altair~=4.0.0",
        "altair-transform",
        "ibis-framework~=1.3.0",
        "jaeger-client",
        "jupyter_jaeger~=1.0.3",
        "jupyterlab<=2",
        "mypy_extensions",
        "pymapd",
        "tornado",
        "typing_extensions",
        "vega_datasets",
    ],
    extras_require={"dev": ["black"]},
    python_requires=">=3.6",
    include_package_data=True,
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
